"""
Umeboshi provides two locking mechanisms. If using a cache with its own
`lock()` method (such as [python redis lock][prs]), it will make use of that.
Otherwise it provides its own simple locking implementation.

[prs]: https://pypi.python.org/pypi/python-redis-lock
"""
from django.core.cache import cache
import contextlib
import time

__all__ = ('lock',)

lock_key = 'umeboshi-event-{}'


@contextlib.contextmanager
def simple_lock(lock_key, timeout=5000):
    """
    A simple context manager that raises the passed exception
    if a lock can't be acquired.
    """

    acquire_lock = lambda: cache.add(lock_key, 1, timeout)
    release_lock = lambda: cache.delete(lock_key)

    waited, hops = 0, 10
    while not acquire_lock():
        time.sleep(float(hops) / 1000.0)
        waited += hops
        if waited > timeout:
            raise RuntimeError('Lock could not be acquired after {}ms'.format(waited))

    try:
        yield
    finally:
        release_lock()


if hasattr(cache, 'lock'):
    lock = lambda id: cache.lock(lock_key.format(id), expire=15)
else:
    lock = lambda id: simple_lock(lock_key.format(id))



# -*- coding: utf-8 -*-
import logging

from django.utils import timezone
from umeboshi.exceptions import DuplicateEvent
from umeboshi.models import Event
from umeboshi.triggers import TriggerBehavior


logger = logging.getLogger('django-umeboshi')


@classmethod
def schedule(cls, datetime_scheduled=None, args=None, silent=True, **kwargs):
    """
    Schedule a Routine class to be run in the future.
    """
    if args is None:
        args = []
    marshaled_data = Event.marshal_data(args)
    hashed_data = Event.hash_data(marshaled_data)

    # The *_ONCE Trigger Behaviors check for existing Events of the
    # same kind in different states, depending on the Trigger Behavior
    # specified. If the event has a `task_group`, then that will be the
    # kind of of Event to look for; otherwise the trigger name will be
    # used. The developer can give multiple Routines the same task name
    # (and different trigger names, of course), checking for the
    # existence of any of a group of Routines.
    group_check = {"task_group": cls.task_group} if cls.task_group is not None \
        else {"trigger_name": cls.trigger_name}
    behavior = cls.behavior if cls.behavior in TriggerBehavior.values.keys() else TriggerBehavior.DEFAULT

    try:
        if behavior == TriggerBehavior.RUN_AND_SCHEDULE_ONCE \
            and Event.objects.filter(data_hash=hashed_data,
                                     status__in=(Event.Status.SUCCESSFUL, Event.Status.CREATED),
                                     **group_check).exists():
            raise DuplicateEvent()

        if behavior == TriggerBehavior.RUN_ONCE \
            and Event.objects.filter(data_hash=hashed_data,
                                     status=Event.Status.SUCCESSFUL,
                                     **group_check).exists():
            raise DuplicateEvent()

        if behavior == TriggerBehavior.SCHEDULE_ONCE \
            and Event.objects.filter(data_hash=hashed_data,
                                     datetime_processed__isnull=True,
                                     **group_check).exists():
            raise DuplicateEvent()
        if behavior == TriggerBehavior.LAST_ONLY:
            waiting_events = Event.objects.filter(data_hash=hashed_data,
                                                  datetime_processed__isnull=True,
                                                  **group_check)
            for waiting_event in waiting_events:
                waiting_event.cancel()

    except DuplicateEvent:
        if not silent:
            raise
        else:
            return None

    if datetime_scheduled is None:
        datetime_scheduled = timezone.now()

    event = Event.objects.create(trigger_name=cls.trigger_name,
                                 task_group=cls.task_group,
                                 datetime_scheduled=datetime_scheduled,
                                 status=Event.Status.CREATED,
                                 args=args)
    return event


def routine(trigger_name=None, task_group=None, behavior=None):
    """
    Routines are declared with the `routine` decorator, which takes care of
    registering the Routine class with Umeboshi.
    """
    from umeboshi.runner import runner

    def wrapper(cls):
        cls.trigger_name = getattr(cls, 'trigger_name', None) or trigger_name
        cls.task_group = getattr(cls, 'task_group', None) or task_group
        cls.behavior = getattr(cls, 'behavior', None) or behavior
        runner.register(cls)

        cls.schedule = schedule

        return cls

    return wrapper



import logging
from django.utils.module_loading import import_string
from umeboshi.exceptions import UnknownTriggerException, RoutineRunException, NoRoutineTriggerException


class Runner(object):
    """
    The Umeboshi Runner is responsible for registering and retrieving Routines.
    It provides a wrapper around Routine API to establish sane defaults.
    """

    def __init__(self):
        self.registry = {}

    def register(self, cls):
        if cls.trigger_name is None:
            raise NoRoutineTriggerException

        if cls.trigger_name in self.registry:
            self.logger.warning('Duplicate definition for trigger {} at {}.{} and {}.{}',
                                cls.trigger_name, self.register[cls.__module__, cls.trigger_name],
                                cls.__module__, cls.__name__)
        self.registry[cls.trigger_name] = "{}.{}".format(cls.__module__, cls.__name__)

    def get_routine_class(self, trigger_name):
        if trigger_name in self.registry:
            return import_string(self.registry[trigger_name])
            return self.registry[trigger_name]

        raise UnknownTriggerException()

    def check_validity(self, routine):
        if not hasattr(routine, 'check_validity'):
            return True
        return routine.check_validity()

    def run(self, routine):
        if not hasattr(routine, 'run'):
            raise NotImplementedError
        try:
            return routine.run()
        except Exception as e:
            self.logger.exception(e)
            raise RoutineRunException()

    @property
    def logger(self):
        return logging.getLogger('django-umeboshi.runner')

runner = Runner()



import logging

from django.test import TestCase
from django.utils import timezone

from umeboshi import exceptions
from umeboshi.models import Event, TriggerBehavior
from umeboshi.routines import routine


class BaseRoutine(object):

    def __init__(self, args=None):
        self.data = args

    def check_validity(self):
        return True

    def run(self):
        return True

TRIGGER_SIMPLE = 'simple-model'


@routine()
class SimpleRoutine(BaseRoutine):
    trigger_name = TRIGGER_SIMPLE

TRIGGER_RUN_ONCE = 'send_once'


@routine()
class SendOnceRoutine(BaseRoutine):
    trigger_name = TRIGGER_RUN_ONCE
    behavior = TriggerBehavior.RUN_ONCE

SCHEDULE_ONCE = 'trigger_once'


@routine()
class TriggerOnceRoutine(BaseRoutine):
    trigger_name = SCHEDULE_ONCE
    behavior = TriggerBehavior.SCHEDULE_ONCE

RUN_AND_SCHEDULE_ONCE = 'run_and_schedule_once'


@routine()
class RunAndScheduleOnceRoutine(BaseRoutine):
    trigger_name = RUN_AND_SCHEDULE_ONCE
    behavior = TriggerBehavior.RUN_AND_SCHEDULE_ONCE

TRIGGER_DELETE_AFTER_PROCESSING = 'trigger_delete_after'


@routine()
class TriggerDeleteAfterRoutine(BaseRoutine):
    trigger_name = TRIGGER_DELETE_AFTER_PROCESSING
    behavior = TriggerBehavior.DELETE_AFTER_PROCESSING


@routine(trigger_name="base-trigger")
class BaseRoutine(object):
    pass


class ModelTests(TestCase):

    def setUp(self):
        logging.disable(logging.WARNING)

    def test_trigger_event(self):

        welcome_path = "{}.{}".format(SimpleRoutine.__module__, SimpleRoutine.__name__)
        trigger_settings = {TRIGGER_SIMPLE: welcome_path}
        with self.settings(UMEBOSHI_TRIGGERS=trigger_settings):
            event = SimpleRoutine.schedule()

            self.assertEqual(event.trigger_name, TRIGGER_SIMPLE)
            diff = event.datetime_created - event.datetime_scheduled
            self.assertLessEqual(abs(diff.seconds), 1)

    def test_trigger_event_with_data(self):

        welcome_path = "{}.{}".format(SimpleRoutine.__module__, SimpleRoutine.__name__)
        trigger_settings = {TRIGGER_SIMPLE: welcome_path}
        with self.settings(UMEBOSHI_TRIGGERS=trigger_settings):
            data = ['World', 5, timezone.now().date()]
            event = SimpleRoutine.schedule(args=data)

            self.assertEqual(event.trigger_name, TRIGGER_SIMPLE)

            diff = event.datetime_created - event.datetime_scheduled
            self.assertLessEqual(abs(diff.seconds), 1)

            event = Event.objects.get(pk=event.id)
            self.assertEqual(event.args, data)

    def test_trigger_event_run_once(self):

        welcome_path = "{}.{}".format(SendOnceRoutine.__module__, SendOnceRoutine.__name__)
        trigger_settings = {TRIGGER_RUN_ONCE: welcome_path}
        with self.settings(UMEBOSHI_TRIGGERS=trigger_settings):
            data = ['unique_data']
            # First event
            event = SendOnceRoutine.schedule(args=data)
            self.assertEqual(event.trigger_name, TRIGGER_RUN_ONCE)
            event.process()

            # Second event
            with self.assertRaises(exceptions.DuplicateEvent):
                SendOnceRoutine.schedule(args=data, silent=False)

    def test_trigger_event_schedule_once(self):

        welcome_path = "{}.{}".format(TriggerOnceRoutine.__module__, TriggerOnceRoutine.__name__)
        trigger_settings = {SCHEDULE_ONCE: welcome_path}
        with self.settings(UMEBOSHI_TRIGGERS=trigger_settings):
            data = ['unique_data']
            # First event
            event = TriggerOnceRoutine.schedule(args=data)
            self.assertEqual(event.trigger_name, SCHEDULE_ONCE)

            # Second event while 1st is not processed yet
            with self.assertRaises(exceptions.DuplicateEvent):
                TriggerOnceRoutine.schedule(args=data, silent=False)

            # Mark 1st event as processed
            event.datetime_processed = timezone.now()
            event.save()

            # Second event while 1st IS NOW processed
            new_event = TriggerOnceRoutine.schedule(args=data)
            self.assertNotEqual(new_event, event)
            self.assertEqual(new_event.trigger_name, SCHEDULE_ONCE)

    def test_run_and_schedule_event_schedule_once(self):

        welcome_path = "{}.{}".format(RunAndScheduleOnceRoutine.__module__, RunAndScheduleOnceRoutine.__name__)
        trigger_settings = {RUN_AND_SCHEDULE_ONCE: welcome_path}
        with self.settings(UMEBOSHI_TRIGGERS=trigger_settings):
            data = ['unique_data']
            # First event
            event = RunAndScheduleOnceRoutine.schedule(args=data)
            self.assertEqual(event.trigger_name, RUN_AND_SCHEDULE_ONCE)

            # Second event while 1st is not processed yet
            with self.assertRaises(exceptions.DuplicateEvent):
                RunAndScheduleOnceRoutine.schedule(args=data, silent=False)

            # Mark 1st event as processed
            event.datetime_processed = timezone.now()
            event.save()

            # Second event while 1st IS NOW processed
            with self.assertRaises(exceptions.DuplicateEvent):
                RunAndScheduleOnceRoutine.schedule(args=data, silent=False)

    def test_trigger_event_delete_after_processing(self):

        welcome_path = "{}.{}".format(TriggerDeleteAfterRoutine.__module__, TriggerDeleteAfterRoutine.__name__)
        trigger_settings = {TRIGGER_DELETE_AFTER_PROCESSING: welcome_path}
        with self.settings(UMEBOSHI_TRIGGERS=trigger_settings):
            data = ['unique_data']
            # First event
            event = TriggerDeleteAfterRoutine.schedule(args=data)
            self.assertEqual(event.trigger_name, TRIGGER_DELETE_AFTER_PROCESSING)

            # Second event while 1st is not processed yet
            second_event = TriggerDeleteAfterRoutine.schedule(args=data)
            self.assertEqual(second_event.trigger_name, TRIGGER_DELETE_AFTER_PROCESSING)

            # Mark 1st event as processed
            event.process()

            with self.assertRaises(Event.DoesNotExist):
                Event.objects.get(pk=event.id)

            second_event = Event.objects.get(pk=second_event.id)
            self.assertEqual(second_event.status, Event.Status.CREATED)

    def test_base_trigger(self):
        self.assertEquals(BaseRoutine.trigger_name, "base-trigger")
        self.assertIsNone(BaseRoutine.task_group)
        self.assertIsNone(BaseRoutine.behavior)

        event = BaseRoutine.schedule()
        with self.assertRaises(NotImplementedError):
            event.process()

        BaseRoutine.run = lambda self: None
        event.process()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django_extensions.db.fields


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Event',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('uuid', django_extensions.db.fields.ShortUUIDField(unique=True, editable=False, blank=True)),
                ('trigger_name', models.CharField(max_length=50, db_index=True)),
                ('task_group', models.CharField(max_length=256, null=True, db_index=True)),
                ('data_pickled', models.TextField(editable=False, blank=True)),
                ('data_hash', models.CharField(max_length=32, db_index=True)),
                ('datetime_created', models.DateTimeField(auto_now_add=True, null=True)),
                ('datetime_scheduled', models.DateTimeField(db_index=True)),
                ('datetime_processed', models.DateTimeField(null=True, db_index=True)),
                ('status', models.IntegerField(default=-3, db_index=True)),
            ],
        ),
        migrations.AlterIndexTogether(
            name='event',
            index_together=set([('datetime_processed', 'datetime_scheduled'), ('data_hash', 'datetime_processed', 'trigger_name')]),
        ),
    ]



from django_enumfield import enum


class TriggerBehavior(enum.Enum):

    """
    Trigger Behaviors govern when to allow an Event to be scheduled.
    """
    DEFAULT = "default"

    # Disallow if there is already an event with this name/data waiting to be
    # processed.
    SCHEDULE_ONCE = "schedule-once"

    # Disallow if an event with this name/data has run successfully.
    RUN_ONCE = "run-once"

    # Disallow if an event with this name/data has run successfully, or is
    # scheduled to run.
    RUN_AND_SCHEDULE_ONCE = "run-and-schedule-once"

    # Cancel any waiting events and schedule this one instead.
    LAST_ONLY = "last-only"

    # Delete event after processing
    DELETE_AFTER_PROCESSING = "delete-after-processing"



# -*- coding: utf-8 -*-
"""
django-umeboshi.tasks
~~~~~~~~~~~~~~~~~~~~~~~~~~

Tasks to run asynchronously via Celery
"""

from umeboshi.models import Event
from django.utils import timezone
from celery.task import task
from umeboshi.utils import lock


@task(ignore_result=True, soft_time_limit=300)
def process_event(event_id):
    with lock(event_id):
        # Load event
        event = Event.objects.get(pk=event_id)

        # Process if not processed already
        if event.status == Event.Status.CREATED:
            event.process()


@task(ignore_result=True)
def process_all_events():
    event_ids = Event.objects.filter(datetime_scheduled__lte=timezone.now(),
                                     status=Event.Status.CREATED).order_by('datetime_scheduled')\
        .values_list('pk', flat=True)

    for event_id in event_ids:
        process_event.delay(event_id)

    return len(event_ids)



import django

__doc__ = """Minimal django settings to run manage.py test command"""

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': __name__,
    }
}

BROKER_BACKEND = 'memory'

ROOT_URLCONF = 'tests.urls'

from celery import current_app

CELERY_ALWAYS_EAGER = True
CELERY_EAGER_PROPAGATES_EXCEPTIONS = True
current_app.conf.CELERY_ALWAYS_EAGER = True
current_app.conf.CELERY_EAGER_PROPAGATES_EXCEPTIONS = True

CELERY_EAGER_TRANSACTION = True

MIDDLEWARE_CLASSES = ()

INSTALLED_APPS = ('django.contrib.auth',
                  'django.contrib.contenttypes',
                  'umeboshi',
                  'test'
)

USE_TZ = True
SECRET_KEY = "django_tests_secret_key"
TIME_ZONE = 'America/New_York'
LANGUAGE_CODE = 'en-us'
ADMIN_MEDIA_PREFIX = '/static/admin/'
STATICFILES_DIRS = ()

if django.VERSION < (1, 6):
    TEST_RUNNER = 'django.test.simple.DjangoTestSuiteRunner'
else:
    TEST_RUNNER = 'django.test.runner.DiscoverRunner'



from django.contrib import admin
from umeboshi.models import Event


def link_to_field(field_name, short_description=None, getter=None):
    """
    Generic field linking for to-one.
    """

    def fn(obj):
        field = getter(obj) if getter else getattr(obj, field_name)

        if field:
            return '<a href="{}">{}</a>'.format(field.get_admin_url(), field)
        return '(None)'

    fn.short_description = short_description or field_name.title()
    fn.admin_order_field = field_name
    fn.allow_tags = True
    return fn


class EventAdmin(admin.ModelAdmin):

    list_filter = ['trigger_name', 'status']

    date_hierarchy = 'datetime_processed'

    list_display = [
        'id',
        'trigger_name',
        'status',
        'datetime_created',
        'datetime_scheduled',
        'datetime_processed'
    ]

    readonly_fields = ['id', 'trigger_name', 'datetime_created',
                       'datetime_processed', '_args']

    fieldsets = (('Events',
                  {'fields':
                   ('id',
                    'trigger_name',
                    '_args')}),
                 ('Status',
                  {'fields':
                   ('status', 'datetime_created', 'datetime_scheduled', 'datetime_processed')}))

    def _args(self, obj):
        return ', '.join(str(a) for a in obj.args)

admin.site.register(Event, EventAdmin)



# coding=utf-8
import os
import sys

from setuptools import setup, Command, find_packages



class RunTests(Command):

    """RunTests class borrowed from django-celery project
    """
    description = 'Run the django test suite from the tests dir.'
    user_options = []
    extra_args = []

    def run(self):
        from django.core.management import execute_from_command_line
        settings_module_name = 'tests.settings'
        os.environ['DJANGO_SETTINGS_MODULE'] = os.environ.get(
            'DJANGO_SETTINGS_MODULE',
            settings_module_name)
        prev_argv = sys.argv[:]

        this_dir = os.getcwd()
        testproj_dir = os.path.join(this_dir, 'tests')
        os.chdir(testproj_dir)
        sys.path.append(testproj_dir)

        try:
            sys.argv = [__file__, 'test'] + self.extra_args
            execute_from_command_line(argv=sys.argv)
        finally:
            sys.argv[:] = prev_argv

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass
setup(
    name="django-umeboshi",
    version="0.4",
    author="MakeSpace Labs, Inc.",
    author_email="zach.smith@makespace.com",
    description="Umeboshi is a Django application for durable long-term scheduling of arbitrary actions.",
    url="https://github.com/makingspace/umeboshi",
    license="Revised BSD",
    packages=find_packages(),
    install_requires=[
        "celery>=2.2.7",
        "Django>=1.8",
        "django_extensions",
        "django_enumfield",
        "shortuuid",
        "django-model-utils"
    ],
    classifiers=[
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Database",
    ],
    cmdclass={'test': RunTests},
)



# -*- coding: utf-8 -*-
"""
django-umeboshi.exceptions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains all exceptions used by the Django Umeboshi module
"""


class NoRoutineTriggerException(Exception):

    """ No trigger_name was provided when registering a Routine """


class DuplicateEvent(Exception):

    """ Event could not be triggered again because of trigger limitation """


class RoutineRunException(Exception):

    """ An error occurred during the run() method of Event processing """


class UnknownTriggerException(Exception):

    """ Event's trigger is not defined """


class RoutineRetryException(Exception):

    """ Could not complete; schedule for later """

    def __init__(self, new_datetime=None):
        # The exception is raised with a new datetime, which is then inspected
        # by the Event's `process()` method for scheduling a new Event.
        self.new_datetime = new_datetime



# -*- coding: utf-8 -*-
"""
django-umeboshi.models
~~~~~~~~~~~~~~~~~~~~~~~~~~~

Umeboshi uses the Event model to repesent a single instance of deferred
computation. It is represented as a reference to a specific class defined
in the application (a Routine), combined with the arguments passed to that
Routine's `_run` function and the details of its scheduling. This includes,
after the computation is processed, the status of the computation.
"""
import hashlib
import pickle

from django.core.exceptions import ValidationError
from django.core.urlresolvers import reverse
from django.db import models
from django.utils import timezone
from django.utils.timezone import now, timedelta
from django_enumfield import enum
from django_extensions.db import fields
from model_utils.managers import QueryManager

from umeboshi.exceptions import RoutineRunException, RoutineRetryException
from umeboshi.triggers import TriggerBehavior


class BaseModel(models.Model):

    class Meta:
        abstract = True

    def get_admin_url(self):
        def view_name_for_model(model):
            return "admin:{}_{}_change".format(model._meta.app_label, model._meta.model_name)

        return reverse(view_name_for_model(self), args=(self.id,))

    def instance_from_db(self):
        """
        Returns a fresh copy of this object from db.
        """
        return self.__class__.objects.get(pk=self.pk)


class EventManager(QueryManager):

    def get_routine_events(self, routine):
        """
        Return all Events of a given type that are currently scheduled.
        """
        return self.filter(trigger_name=routine.trigger_name,
                           status=Event.Status.CREATED).order_by('datetime_scheduled')


class Event(BaseModel):

    """
    Events are the way that Umeboshi Routines are scheduled to be run. An Event
    object is saved to the database with the `trigger_name` of the Routine, as
    well as the pickled arguments to the Routine and the scheduling details.
    """

    class Meta:
        app_label = 'umeboshi'
        index_together = (
            ('datetime_processed', 'datetime_scheduled'),
            ('data_hash', 'datetime_processed', 'trigger_name')
        )

    uuid = fields.ShortUUIDField(unique=True, editable=False)
    objects = EventManager()
    # The trigger name corresponds to the name given a Umeboshi Routine with the
    # `scheduled` decorator in the application logic.
    trigger_name = models.CharField(db_index=True, max_length=50)
    task_group = models.CharField(db_index=True, max_length=256, null=True)
    data_pickled = models.BinaryField(blank=True, editable=False)
    data_hash = models.CharField(db_index=True, max_length=32)
    datetime_created = models.DateTimeField(null=True, auto_now_add=True)
    datetime_scheduled = models.DateTimeField(db_index=True)
    datetime_processed = models.DateTimeField(db_index=True, null=True)

    class Status(enum.Enum):

        """
        Event statuses are stored with the object. An Event scheduled to be
        processed in the future is `CREATED`; after processing it can be in a
        variety of states.
        """
        CREATED = 0
        # If an exception is raised during the main task body of an Event's
        # Routine, it will be marked `FAILED`.
        FAILED = -1
        # If an Event is cancelled beforehand, or if its validity check fails
        # during processing, it will be marked `CANCELLED`.
        CANCELLED = -2
        # If an Event fails anywhere else during processing (for instance, in
        # its Routine's `__init__` method), it will be marked `BROKEN`.
        BROKEN = -3
        # Finally, after successful processing, an Event will be marked
        # `SUCCESSFUL`.
        SUCCESSFUL = 1

    status = enum.EnumField(Status, default=Status.CREATED)

    @staticmethod
    def marshal_data(data):
        """
        Events use `pickle` to marshal their argument data for storage.
        """
        return pickle.dumps(data)

    @staticmethod
    def unmarshal_data(data):
        return pickle.loads(data)

    @staticmethod
    def hash_data(data):
        """
        Umeboshi calculates an md5 hash of argument data.
        """
        return hashlib.md5(data).hexdigest()

    @property
    def args(self):
        """
        Unmarshaled data is available for inspection on the instantiated Event.
        """
        if not hasattr(self, '_data'):
            self._data = [] if self.has_data \
                else self.unmarshal_data(self.data_pickled)
        return self._data

    @args.setter
    def args(self, value):
        self._data = value

    @property
    def has_data(self):
        return len(self.data_pickled) <= 0

    def process(self):
        """
        When an Event's scheduled datetime comes up, it will be processed.
        """
        from umeboshi.runner import runner
        try:
            # The class is retrieved according to the trigger name.
            routine_class = runner.get_routine_class(self.trigger_name)
            # The routine is then instantiated with the unmarshaled data that
            # had been saved with the Event.
            routine = routine_class(*self.args)
            # Before running the Routine, the Event will run the Routine's
            # validity check. This allows the Routine to specify conditions
            # that must be met at runtime (as opposed to when the Event is
            # scheduled) for the Routine to be run.
            if not runner.check_validity(routine):
                self.status = self.Status.CANCELLED
            else:
                # If the routine is still valid, it will be run and the Event
                # will be marked `SUCCESSFUL`.
                runner.run(routine)
                self.status = self.Status.SUCCESSFUL
            if routine_class.behavior == TriggerBehavior.DELETE_AFTER_PROCESSING:
                self.delete()
        except RoutineRunException:
            self.status = self.Status.FAILED
        except RoutineRetryException as e:
            self.status = self.Status.FAILED
            if e.new_datetime:
                self.retry_schedule(e.new_datetime)
            else:
                self.retry_schedule()
        except:
            self.status = self.Status.BROKEN
            raise
        finally:
            if self.pk:
                self.datetime_processed = timezone.now()
                self.save()

    def retry_schedule(self, new_datetime=now() + timedelta(hours=1)):
        if self.status in (self.Status.SUCCESSFUL, self.Status.CREATED):
            raise ValidationError("Can only reschedule a failed event.")
        else:
            return Event.objects.create(
                trigger_name=self.trigger_name,
                datetime_scheduled=new_datetime,
                status=Event.Status.CREATED,
                args=self.args
            )

    def cancel(self):
        self.datetime_processed = timezone.now()
        self.status = self.Status.CANCELLED
        self.save()

    def save(self, *args, **kwargs):
        # The Event's arguments are marshaled before it's saved to the db.
        self.data_pickled = self.marshal_data(self.args)
        self.data_hash = self.hash_data(self.data_pickled)
        super(Event, self).save(*args, **kwargs)

    def __unicode__(self):
        return 'Umeboshi Event #{}'.format(self.pk)
